export namespace session {
    /**
     * Get the current session operations.
     * @returns {SessionInterface}
     */
    function getOperations(): SessionInterface;
    /**
     * Customize your session operations.
     * @param {SessionInterface} operations
     */
    function setOperations(operations: SessionInterface): Promise<void>;
    /**
     * Start a parked session from `cookies` or create a new one if no
     * parked session is found or is expired.\
     * @param {StartPayload} payload
     * @returns {Promise<import('sveltekit-unsafe').Unsafe<import('./types').Session>>}
     */
    function start({ cookies }: StartPayload): Promise<import("sveltekit-unsafe").Unsafe<import("./types").Session>>;
    /**
     * Destroy a session.
     * @param {{id:string}} payload
     * @returns {Promise<import('sveltekit-unsafe').Unsafe<void>>} Error if no session with the given `id` is found, otherwise success.
     */
    function destroy({ id }: {
        id: string;
    }): Promise<import("sveltekit-unsafe").Unsafe<void>>;
    /**
     * Clear all invalid local sessions.
     * - You should invoke this periodically.
     * - This uses an internal semaphore and too many `flush()` submissions will be ignored silently.
     */
    function flush(): Promise<void>;
}
export type StartPayload = {
    /**
     * Request cookies.\
     * Usually retrieved from the SvelteKit endpoint definition itself, for example
     * ```js
     * export function GET({ cookies }) {
     * session.start({ cookies })
     * return new Response('hello')
     * }
     * ```
     */
    cookies: import('@sveltejs/kit').Cookies;
};
export type FindPayload = {
    /**
     * Session id.
     */
    id: string;
};
export type CreatePayload = {
    id: string;
    data: any;
};
export type Exists = (id: string) => Promise<import('sveltekit-unsafe').Unsafe<boolean>>;
export type IsValid = (id: string) => Promise<import('sveltekit-unsafe').Unsafe<boolean>>;
export type Has = (id: string) => Promise<import('sveltekit-unsafe').Unsafe<boolean>>;
export type Get = (id: string) => Promise<import('sveltekit-unsafe').Unsafe<import('./types').Session>>;
export type Delete = (id: string) => Promise<import('sveltekit-unsafe').Unsafe<void>>;
export type Set = (id: string, session: import('./types').Session) => Promise<import('sveltekit-unsafe').Unsafe<void>>;
export type SessionInterface = {
    /**
     * Check wether or not the session id already exists.
     */
    exists: Exists;
    /**
     * Check wether or not the session id is valid.\
     * This is where you may checked wether or not the session has expired or not, for example.
     */
    isValid: IsValid;
    /**
     * Get the session.
     */
    has: Has;
    /**
     * Get the session.
     */
    get: Get;
    /**
     * Set a session.
     */
    set: Set;
    /**
     * Delete a session.
     */
    delete: Delete;
};
